// JoyCon Test  version 0.2   Copyright(c)2018 by E.Kako  
//   This version supports JoyCon-R only.

#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#define FUNC __declspec(dllimport) __stdcall
extern "C" HANDLE FUNC OpenHidHandle(unsigned short vendor_id, unsigned short product_id);
extern "C" void FUNC ReadReport(HANDLE handle,unsigned char *InputReport,int *len);
extern "C" void FUNC WriteReport(HANDLE handle,unsigned char *OutputReport, int *len);
extern "C" void FUNC CloseHidHandle(HANDLE handle);

HANDLE hRsDevHandle;

const USHORT VID = 0x057e; // Nintendo
const USHORT PID = 0x2007; // JonCon - R

#define MAXREPORTSIZE 256
int		InputLength,OutputLength;
unsigned char	InputReport[MAXREPORTSIZE];
unsigned char	OutputReport[MAXREPORTSIZE];
int packet_num = 0;

unsigned char	joycon_loop_flag;
unsigned char	joycon_in_use;
uintptr_t joycon_th;

unsigned char	res_type;
unsigned char	button1,button2;
unsigned int	StickX,StickY;
unsigned int	Ax,Ay,Az;
unsigned int	Rx,Ry,Rz;

double tune_freq[60] = {
	55,58.27,61.735,65.406,69.296,73.416,77.782,82.407,87.307,92.499,97.999,103.826,
	110,116.541,123.471,130.813,138.591,146.832,155.563,164.814,174.614,184.997,195.998,207.652,
	220,233.082,246.942,261.626,277.183,293.665,311.127,329.628,349.228,369.994,391.995,415.305,
	440,466.164,493.883,523.251,554.365,587.330,622.254,659.255,698.456,739.989,783.911,830.609,
	880,932.328,987.767,1046.50,1108.73,1174.65,1244.50,1318.51,1396.91,1479.97,1567.98,1661.21,
};
int freq_param1[60];
int freq_param2[60];

unsigned char amp1[11] = { 0x00,0x1c,0x38,0x58,0x74,0x88,0x98,0xa8,0xb4,0xc0,0xc8 };
unsigned char amp2[11] = { 0x40,0x47,0x4e,0x56,0x5d,0x62,0x66,0x6a,0x6d,0x70,0x72 };

int a1[60] = {
	0 ,0 ,0, 0 ,0 , 0, 0, 0, 0, 0 ,
	0 ,0 ,10,10,9 , 8, 8, 7, 5, 3 ,
	2, 3 ,4, 5 ,6 , 7, 7, 7, 7, 7 ,
	7 ,4 ,4, 9 ,9 , 9, 8, 10,10,10,
	10,10,10,10,10, 10,10,10,10,10,
	10,10,10,10,10, 10,10,10,10,10,
};
int a2[60] = {
	0 ,0 ,0, 0 ,0 , 0, 0, 0, 0, 0 ,
	0 ,0 ,10,10,9 , 8, 8, 7, 5, 3 ,
	2 ,3 ,4, 5 ,6 , 7, 7, 7, 7, 7 ,
	7 ,4 ,4, 9 ,9 , 9, 8, 10,10,10,
	10,10,10,10,10, 10,10,10,10,10,
	10,10,10,10,10, 10,10,10,10,10,
};

void Sleep2(int msec)
{
	LARGE_INTEGER f; // for hi-res. performance counter
	LARGE_INTEGER t1,t2; // for hi-res. performance counter
	double t;
	QueryPerformanceFrequency(&f);
	QueryPerformanceCounter(&t1);
	while (1) {
		QueryPerformanceCounter(&t2);
		t = (double)t2.QuadPart - (double)t1.QuadPart;
		t /= (double)f.QuadPart;
		t *= 1000;
		if (t>(double)msec) { return; }
	}
}

void cmd_0x01_parameter_set()
{
	OutputReport[0]=0x01;
	OutputReport[1]=( (++packet_num) % 16);
	OutputReport[2]=0; OutputReport[3]=0; OutputReport[4]=0; OutputReport[5]=0;
	OutputReport[6]=0; OutputReport[7]=0; OutputReport[8]=0; OutputReport[9]=0;
}

void Report_subcmd_0x30(unsigned char data)
{
	cmd_0x01_parameter_set();
	OutputReport[10]= 0x30;
	OutputReport[11]= data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
	ReadReport( hRsDevHandle , InputReport, &InputLength);
}

void Report_subcmd_0x03(unsigned char data)
{
	cmd_0x01_parameter_set();
	OutputReport[10]= 0x03;
	OutputReport[11]= data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
	ReadReport( hRsDevHandle , InputReport, &InputLength);
}

void Report_subcmd_0x40(unsigned char data)
{
	cmd_0x01_parameter_set();
	OutputReport[10]= 0x40;
	OutputReport[11]= data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
	ReadReport( hRsDevHandle , InputReport, &InputLength);
}

void Report_subcmd_0x48(unsigned char data)
{
	cmd_0x01_parameter_set();
	OutputReport[10]= 0x48;
	OutputReport[11]= data;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
	ReadReport( hRsDevHandle , InputReport, &InputLength);
}

void Joycon_mode_0x30(void)
{
	Report_subcmd_0x03(0x30); // standard full mode
	// respond automatically (interval = ??)
}

void Joycon_mode_0x3F(void)
{
	Report_subcmd_0x03(0x3F); // button ONLY Report Mode (HID gamepad mode)
	// respond when button changes only
}

void Joycon_IMU_enable(void)
{
	Report_subcmd_0x40(0x01);
}

void Joycon_IMU_disable(void)
{
	Report_subcmd_0x40(0x00);
}

void Joycon_LED_on()
{
	Report_subcmd_0x30(0x01);  // set LED data  0x01 ... LED1 on
	// 0x01 = led 1(top)   ,   0x02 = led 2        ,  0x04 = led 3        ,  0x08 = led 4
	// 0x10 = led 1 blink  ,   0x20 = led 2 blink  ,  0x40 = led 3 blink  ,  0x80 = led 4 blink
}

void Joycon_LED_off()
{
	Report_subcmd_0x30(0x00);  // set LED data -> all OFF
}

void Joycon_Haptic_enable(void)
{
	Report_subcmd_0x48(0x01);
}

void Joycon_Haptic_disable(void)
{
	Report_subcmd_0x48(0x00);
}

void Joycon_Haptic_test(unsigned char d1,unsigned char d2,unsigned char d3,unsigned char d4)
{
	OutputReport[0]=0x10;
	OutputReport[1]=( (++packet_num) % 16);
	OutputReport[2]=d1; OutputReport[3]=d2; OutputReport[4]=d3; OutputReport[5]=d4;
	OutputReport[6]=d1; OutputReport[7]=d2; OutputReport[8]=d3; OutputReport[9]=d4;
	WriteReport( hRsDevHandle , OutputReport , &OutputLength);
}

void Joycon_Haptic_off()
{
	Joycon_Haptic_test(0x3c,0,0x2f,0x40);
}

void Joycon_Input(void)
{
	int l;
	ReadReport( hRsDevHandle , InputReport, &InputLength);
	res_type=InputReport[0];
	if (InputLength == 362) {
	//	button1=InputReport[5]; button2=InputReport[4]; // JoyCon-L
		button1=InputReport[3]; button2=InputReport[4]; // JoyCon-R
	//	StickX = InputReport[6] | ((unsigned int)InputReport[7]&0x0F)<<8;  // JoyCon-L
	//	StickY = (InputReport[7]>>4 ) | (InputReport[8] <<4 );
		StickX = InputReport[9] | ((unsigned int)InputReport[10]&0x0F)<<8;  // JoyCon-R
		StickY = (InputReport[10]>>4 ) | (InputReport[11] <<4 );
		Ax=InputReport[13] | (unsigned int)InputReport[14]<<8;
		Ay=InputReport[15] | (unsigned int)InputReport[16]<<8;
		Az=InputReport[17] | (unsigned int)InputReport[18]<<8;
		Rx=InputReport[19] | (unsigned int)InputReport[20]<<8;
		Ry=InputReport[21] | (unsigned int)InputReport[22]<<8;
		Rz=InputReport[23] | (unsigned int)InputReport[24]<<8;
	} else {
		printf("%d\n",InputLength);
	}
}

void Joycon_Input_Main(void *)
{
	// Joycon input thread
	while (joycon_loop_flag) {
		joycon_in_use=1;
		Joycon_Input();
		Sleep2(16);
	}
	joycon_in_use=0;
	return;
}

void Start_Joycon_th(void)
{
	joycon_loop_flag=1;
	joycon_th = _beginthread( Joycon_Input_Main , 0 , NULL );
}

int Joycon_th_check(void)
{
	res_type=0;
	Sleep2(100);
	return (res_type!=0);
}

void Stop_Joycon_th(void)
{
	joycon_loop_flag=0;
	Sleep2(100);
	if (joycon_in_use==1) {
		TerminateThread( (HANDLE)joycon_th , -1); // kill thread  if it stucks
	}
}

void MainLoop()
{
	unsigned char c;
	while (1) {
		if (kbhit()) { c= getch();
			if (c==0x1B) { break; }
		}
		printf("button=%02X,%02X , stick=%04X,%04X , Acc=%04X,%04X,%04X , Gyr=%04X,%04X,%04X\n",
			button1,button2,StickX,StickY,Ax,Ay,Az,Rx,Ry,Rz);
		Sleep2(16);
	}
}

void freq_parameter_calculation()
{
	int i,p0,p1,p2;
	for (i=0;i<60;i++) {
		p0=log2(tune_freq[i]/10.0)*32.0;
		p1=p0-96;
		if (p1<0) { p1=0; } else if (p1>127) { p1=0; }
		p2=p0-64; 
		if (p2<0) { p2=0; } else if (p2>127) { p2=0; }
		freq_param1[i]=p1*4;
		freq_param2[i]=p2;
	}
}

void play_haptic(int note,int length,int tempo)
{
	int i;
	for (i=0;i<length;i++) {
		if (note>=0) {
			if (i==length-1) {
				Joycon_Haptic_test(
					freq_param1[note]&0xFF,
					(freq_param1[note]>>8)+amp1[a1[note]],
					freq_param2[note],
					amp2[a2[note]] );
				Sleep2(tempo/2);
				Joycon_Haptic_test(
					freq_param1[note]&0xFF,
					(freq_param1[note]>>8)+amp1[a1[note]/2],
					freq_param2[note],
					amp2[a2[note]/2] );
				Sleep2(tempo/2);
			} else {
				Joycon_Haptic_test(
					freq_param1[note]&0xFF,
					(freq_param1[note]>>8)+amp1[a1[note]],
					freq_param2[note],
					amp2[a2[note]] );
				Sleep2(tempo);
			}
		} else {
			Sleep2(tempo);
		}
	}
}

void play(char *s)
{
	// this function has too few error checks...
	static int tempo=100;
	static int length=4;
	static int octave=4;
	int res,ln,oc,n,note;
	//printf("play data = \"%s\"\n",s);
	switch (s[0]) {
		case 't': case 'T':
			res=sscanf(s+1,"%d",&tempo);
			if (tempo<=0) { tempo=100; }
			break;
		case 'l': case 'L':
			res=sscanf(s+1,"%d",&length);
			if (length<=0) { length=1; }
			if (length>16) { length=16; }
			break;
		case 'o': case 'O':
			res=sscanf(s+1,"%d",&octave);
			if (octave<1) { octave=1; }
			if (octave>5) { octave=5; }
			break;
		case '<':
			octave--;
			if (octave<1) { octave=1; }
			break;
		case '>':
			octave++;
			if (octave>5) { octave=5; }
			break;
		case 'r': case 'R':
			res=sscanf(s+1,"%d",&ln);
			if (res<=0) {
				play_haptic(-1,length,tempo);
			} else {
				play_haptic(-1,ln,tempo);
			}
			break;
		case 'c': case 'C': case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
		case 'g': case 'G': case 'h': case 'H': case 'a': case 'A': case 'b': case 'B':
			n=tolower(s[0])-'a';
			n-=2; if (n<0) { n+=7; }
			if      (n==0) { note=3; } // C
			else if (n==1) { note=5; } // D
			else if (n==2) { note=7; } // E
			else if (n==3) { note=8; } // F
			else if (n==4) { note=10; } // G
			else if (n==5) { note=12; } // A
			else if (n==6) { note=14; } // B
			note+=(octave-1)*12;
			s++;
			if (s[0]=='+') {
				note++; s++;
			} else if (s[0]=='-') {
				note--; s++;
			}
			if (note<0) { note=-1;}
			if (note>59) { note=-1;}
			res=sscanf(s,"%d",&ln);
			if (res<=0) {
				play_haptic(note,length,tempo);
			} else {
				play_haptic(note,ln,tempo);
			}
			break;
	}
}

void play_music()
{
	freq_parameter_calculation(); // init
	play("t10"); play("l2"); play("o4");
	play("<"); play("b"); play(">"); play("d"); play("g"); play("b"); play("g6");
}

int main( int argc, char *argv[])
{
	printf("JoyCon test ver 0.2 by E.Kako.\n");

	hRsDevHandle = OpenHidHandle(VID,PID);
	if ( hRsDevHandle == INVALID_HANDLE_VALUE) {
		MessageBox(0,"JoyCon not found.",0,0);
		return 1;
	}

	Joycon_mode_0x3F(); // HID mode
	Joycon_LED_on();

	Joycon_Haptic_enable();
	play_music();
	Joycon_Haptic_off();
	Joycon_Haptic_disable();

	Joycon_mode_0x30(); // joycon standard mode
	Joycon_IMU_enable(); // gyro acc sensor enable
	Joycon_LED_on();
	Start_Joycon_th();
	if ( Joycon_th_check() ) {
		MainLoop();
	} else {
		MessageBox(0,"Joycon-R cannot open.",0,0);
	}
	Stop_Joycon_th();
	Joycon_LED_off();
	Joycon_IMU_disable();
	Joycon_mode_0x3F(); // HID mode
	CloseHidHandle( hRsDevHandle );
	return 0;
} 
